#ifndef contain_t
#define contain_t

#include <string>
#include "error.h"
#include "funcs.h"
#include "mathops.h"

extern bool debug;

/////////////////
// DEFINITIONS //
/////////////////

/********/
/* list */
/********/

template<class T> class list {

  public:

    // constructors:

    // standard constructor:
    list():
      size(0),
      start(new T[0]) {
    };

    // general constructors:
    list(const unsigned long &n):
      size(n),
      start(new T[n]) {
      init(T(0));
    };
    list(const unsigned long &n, const T& a):
      size(n),
      start(new T[n]) {
      init(a);
    };

    // copy constructor:
    list(const list<T> &l):
      size(l.size),
      start(new T[size]) {
      for (unsigned long int i = 0; i < size; i++)
        start[i] = l.start[i];
    }

    // destructor:
    virtual ~list() {
      delete[] start;
    }

    // internal operators:
    list<T>& operator+=(const list  <T>    &)      ;
          T& operator[](const unsigned long&)      ;
    const T& operator[](const unsigned long&) const;
    list<T>& operator= (const list  <T>    &)      ;
    list<T>& operator= (const vector<T>    &)      ;

    //methods:
    list<T>&      append(const T&, unsigned long n = 1);
    list<T>&      clear (                             );
    list<T>&      erase (const T&, unsigned long n = 0);
    list<T>&      insert(const T&, unsigned long n = 1);
    unsigned long length(                             ) const {return size;};
    list<T>&      remove(          unsigned long      );
    unsigned long seek  (const T&, unsigned long n = 1) const;
    list<T>&      sort  (const char          &d    = 'a',
                               unsigned long first = 1  ,
                               unsigned long last  = 0  );

  protected:
    unsigned long size;
    T *start;

    // method:
    void init(const T &a) {
      for (unsigned int i = 0; i < size; i++)
        start[i] = a;
    }
};

// external operators:
template<class T           > list<T>  operator+ (const list<T >&, const list<T >&);
template<class T1, class T2> bool     operator==(const list<T1>&, const list<T2>&);
template<class T1, class T2> bool     operator!=(const list<T1>&, const list<T2>&);
template<class T           > bool     operator< (const      T  &, const list<T >&);
template<class T           > ostream& operator<<(      ostream &, const list<T >&);



/*******/
/* set */
/*******/

template<class T> class set: public list<T> {

  public:
    // constructors:

    // standard constructor:
    set():
      list<T>() {
    };

    // general constructors:
    set(const list<T>&);

    // copy constructor:
    set(const set<T>&);

    // internal operators:
    set<T>& operator+=(const set <T>&);
    set<T>& operator^=(const set <T>&);
    set<T>& operator-=(const set <T>&);
    set<T>& operator= (const set <T>&);

    //methods:
    set<T>& append  (const T&                       )      ;
    bool    contains(const T&                       ) const;
    set<T>& insert  (const T&, unsigned long n = 1  )      ;
    set<T>& sort    (const char          &d    = 'a',
                           unsigned long first = 1  ,
                           unsigned long last  = 0  )      ;
};

// external operators:
template<class T           > set<T>   operator+ (const set<T >&, const set<T >&);
template<class T           > set<T>   operator^ (const set<T >&, const set<T >&);
template<class T           > set<T>   operator- (const set<T >&, const set<T >&);
template<class T1, class T2> bool     operator==(const set<T1>&, const set<T2>&);
template<class T1, class T2> bool     operator!=(const set<T1>&, const set<T2>&);
template<class T1, class T2> bool     operator< (const set<T1>&, const set<T2>&);
template<class T1, class T2> bool     operator> (const set<T1>&, const set<T2>&);
template<class T1, class T2> bool     operator<=(const set<T1>&, const set<T2>&);
template<class T1, class T2> bool     operator>=(const set<T1>&, const set<T2>&);
template<class T           > ostream& operator<<(      ostream&, const set<T >&);



/**********/
/* vector */
/**********/

template<class T> class vector {
  friend ostream& operator<<(ostream&, const vector<T>&);

  public:
    // constructors:

    // general constructors:
    vector():
      size(0),
      start(new T[0]) {
    }
    vector(const unsigned long &n):
      size(n),
      start(new T[n]) {
      init(T(0));
    };
    vector(const unsigned long &n, const T &ini):
      size(n),
      start(new T[n]) {
      init(ini);
    };
    vector(const list<T> &l):
      size(l.length()),
      start(new T[size]) {
      for (unsigned long i = 0; i < size; i++)
        start[i] = l[i + 1];
    }

    // copy constructor:
    vector(const vector<T> &v):
      size(v.size),
      start(new T[size]) {
      for (unsigned long i = 0; i < size; i++)
        start[i] = v.start[i];
    }

    // destructor:
    ~vector() {
      delete[] start;
    }

    // internal operators:
     vector& operator+=(const vector<T>    &)      ;
     vector& operator-=(const vector<T>    &)      ;
     vector& operator*=(const T            &)      ;
     vector& operator/=(const T            &)      ;
          T& operator[](const unsigned long&)      ;
    const T& operator[](const unsigned long&) const;
     vector& operator =(const vector<T>    &)      ;

    //methods:
    bool          isnull() const             ;
    unsigned long length() const {return size;}

  private:
    unsigned long size ;
    T*            start;

    // methods:
    void init(const T &a) {
      for (unsigned int i = 0; i < size; i++)
        start[i] = a;
    }
};

// external operators:
template<class T> vector<T> operator+ (const vector<T>&                  );
template<class T> vector<T> operator- (const vector<T>&                  );
template<class T> vector<T> operator+ (const vector<T>&, const vector<T>&);
template<class T> vector<T> operator- (const vector<T>&, const vector<T>&);
template<class T> T         operator* (const vector<T>&, const vector<T>&);
template<class T> vector<T> operator* (const vector<T>&, const T        &);
template<class T> vector<T> operator* (const T        &, const vector<T>&);
template<class T> vector<T> operator/ (const vector<T>&, const T        &);
template<class T> bool      operator==(const vector<T>&, const vector<T>&);
template<class T> bool      operator!=(const vector<T>&, const vector<T>&);
template<class T> ostream&  operator<<(      ostream  &, const vector<T>&);



/**********/
/* matrix */
/**********/

template<class T> class matrix {

  public:
    // constructors:

    // general constructors:
    matrix(const unsigned long &n, const T &ini = T(0)):
      height  (n                                 ) ,
      width   (n                                 ) ,
      outstyle("lines"                           ) ,
      Lines   (vector<Line>(n, vector<T>(n, ini))) {
    };
    matrix(const unsigned long &n, const string &one):
      height  (n                                 ) ,
      width   (n                                 ) ,
      outstyle("lines"                           ) ,
      Lines   (vector<Line>(n, vector<T>(n, T(0)))) {
      if (one != "1")
        Error(MatIdNotOne, "matrix()");
      for (unsigned long i = 1; i <= n; i++)
        Lines[i][i] = T(1);
    };
    matrix(const unsigned long &m, const unsigned long &n, const T &ini = T(0)):
      height  (m                                 ) ,
      width   (n                                 ) ,
      outstyle("lines"                           ) ,
      Lines   (vector<Line>(m, vector<T>(n, ini))) {
    };
    matrix(const vector<T> &v):
      height  (v.length()                        ) ,
      width   (1                                 ) ,
      outstyle("lines"                           ) ,
      Lines   (vector<Line>(height, vector<T>(1))) {
      for (unsigned long i = 1; i <= height; i++)
        Lines[i][1] = v[i];
    };

    // internal operators:
          matrix<T>& operator+=(const matrix<T>    &)      ;
          matrix<T>& operator-=(const matrix<T>    &)      ;
          matrix<T>& operator*=(const T            &)      ;
          matrix<T>& operator*=(const matrix<T>    &)      ;
          matrix<T>& operator/=(const T            &)      ;
          matrix<T>& operator^=(const Z            &)      ;
          vector<T>& operator[](const unsigned long&)      ;
    const vector<T>& operator[](const unsigned long&) const;
          matrix<T>& operator =(const matrix<T>    &)      ;
          matrix<T>& operator =(const vector<T>    &)      ;

    //methods:
    unsigned long columns    (                                        ) const;
//    T             det        (                                        ) const;
    string        getoutstyle(                                        ) const;
//    matrix        invert     (                                        )      ;
//    matrix        leaveout   (unsigned long      , unsigned long      ) const;
    unsigned long lines      (                                        ) const;
//    matrix        minor      (unsigned long      , unsigned long,
//                              unsigned long h = 0, unsigned long w = 0) const;
    void          setoutstyle(const string&                           )      ;
    matrix        t          (                                        ) const;
    matrix        transpose  (                                        )      ;

  private:
    typedef vector<T> Line;
    unsigned long height, width;
    string outstyle;
    vector<Line> Lines;
};

// external operators:
template<class T> matrix<T> operator+ (const matrix<T>&                  );
template<class T> matrix<T> operator- (const matrix<T>&                  );
template<class T> matrix<T> operator+ (const matrix<T>&, const matrix<T>&);
template<class T> matrix<T> operator- (const matrix<T>&, const matrix<T>&);
template<class T> matrix<T> operator* (const matrix<T>&, const matrix<T>&);
template<class T> vector<T> operator* (const matrix<T>&, const vector<T>&);
template<class T> vector<T> operator* (const vector<T>&, const matrix<T>&);
template<class T> matrix<T> operator* (const matrix<T>&, const T        &);
template<class T> matrix<T> operator* (const T        &, const matrix<T>&);
template<class T> matrix<T> operator/ (const matrix<T>&, const T        &);
template<class T> bool      operator==(const matrix<T>&, const matrix<T>&);
template<class T> bool      operator!=(const matrix<T>&, const matrix<T>&);
template<class T> ostream&  operator<<(      ostream  &, const matrix<T>&);



///////////////////
//IMPLEMENTATIONS//
///////////////////

/********/
/* list */
/********/

// operators:

// concat operators:

template<class T> list<T> operator+(const list<T> &l1, const list<T> &l2) {
  if (l1.length() == 0)
    return l2;
  if (l2.length() == 0)
    return l1;
  list<T> temp = l1;
  return temp += l2;
}

template<class T> list<T>& list<T>::operator+=(const list<T> &l) {
  if (l.size > 0) {
    unsigned long i = 0;
    list<T> temp = *this;
    delete[] start;
    start = new T[size += l.size];
    for (; i < temp.size; i++)
      start[i] = temp.start[i];
    for (; i < size; i++)
      start[i] = l.start[i - temp.size];
  }
  return *this;
}

// comparison operators:

template<class T1, class T2> bool operator==(const list<T1> &l1, const list<T2> &l2) {
  unsigned long l = l1.length();
  if (l != l2.length())
    return false;
  for (unsigned long i = 1; i <= l; i++)
    if (l1[i] != l2[i])
      return false;
  return true;
}

template<class T1, class T2> bool operator!=(const list<T1> &l1, const list<T2> &l2) {
  return !(l1 == l2);
}

template<class T> bool operator<(const T &e, const list<T> &l) {
  return l.seek(e) ? true : false;
}

// index operators:

template<class T> T& list<T>::operator[](const unsigned long &i) {
  if (i <= 0 || i > size)
    Error(IndexRange, "list[]");
  return start[i - 1];
}

template<class T> const T& list<T>::operator[](const unsigned long &i) const {
  if (i == 0 || i > size)
    Error(IndexRange, "list[]");
  return start[i - 1];
}

// assignment operators:

template<class T> list<T>& list<T>::operator=(const list<T> &l) {
  delete[] start;
  start = new T[size = l.size];
  for (unsigned long i = 0; i < size; i++)
    start[i] = l.start[i];
  return *this;
}

template<class T> list<T>& list<T>::operator=(const vector<T> &v) {
  delete[] start;
  start = new T[size = v.length()];
  for (unsigned long i = 0; i < size; i++)
    start[i] = v[i + 1];
  return *this;
}

// output operator:

template<class T> ostream& operator<<(ostream &out, const list<T> &l) {
  unsigned long len = l.length();
  if (len == 0)
    out << "[]";
  else {
    out << '[' << l[1];
    for (unsigned long i = 2; i <= len; i++)
      out << ", " << l[i];
    out << ']';
  }
  return out;
}


// methods:

template<class T> list<T>& list<T>::append(const T &a, unsigned long n) {
  list<T> temp = *this;
  unsigned long i = 0;
  delete[] start;
  start = new T[size + n];
  for (; i < size; i++)
    start[i] = temp.start[i];
  size += n;
  for (; i < size; i++)
    start[i] = a;
  return *this;
}

template<class T> list<T>& list<T>::clear() {
  size = 0;
  delete[] start;
  start = new T[0];
  return *this;
}

template<class T> list<T>& list<T>::erase(const T &a, const unsigned long n) {
  unsigned long count = 0, i = 0;
  for (; i < size; i++)
    if (start[i] == a)
      count++;
  if (count > 0) {
    list<T> temp = *this;
    if (n == 0 || n > count)
      n = count;
    delete[] start;
    start = new T[size -= n];
    i = 0;
    for (unsigned j = 0; j < temp.size; j++)
      if (temp.start[j] == a && n > 0)
        n--;
      else
        start[i++] = temp.start[j];
  }
  return *this;
}

template<class T> list<T>& list<T>::insert(const T &a, unsigned long n) {
  if (n > 0) {
    list<T> temp = *this;
    unsigned long i = 0;
    if (n > size)
      n = size;
    else
      n--;
    delete[] start;
    start = new T[++size];
    for (; i < n; i++)
      start[i] = temp.start[i];
    start[i++] = a;
    for (; i < size; i++)
      start[i] = temp.start[i - 1];
  }
  return *this;
}

template<class T> list<T>& list<T>::remove(unsigned long n) {
  if (n > 0 && n <= size) {
    list<T> temp = *this;
    unsigned long i = 0;
    n--;
    delete[] start;
    start = new T[--size];
    for (; i < n; i++)
      start[i] = temp.start[i];
    for (; i < size; i++)
      start[i] = temp.start[i + 1];
  }
  return *this;
}

template<class T> unsigned long list<T>::seek(const T &a, unsigned long n) const {
  // returns position of nth occurrence of a in list (0 if not so many)
  // if n is 0: returns number of occurrences of a in list

  unsigned long i = 0;
  if (n == 0) {
    for (; i < size; i++)
      if (start[i] == a)
        n++;
  }
  else {
    for (; i < size; i++)
      if (start[i] == a)
        if (--n == 0) {
          n = i + 1;
          break;
        }
    if (i == size)
      n = 0;
  }
  return n;
}

template<class T> list<T>& list<T>::sort(const char &d, unsigned long first, unsigned long last) {
  // quicksort
  // d = 'a' (ascending) or d = 'd' (descending)

  if (first == 0 || first > size)
    first = 1;
  if (last < first)
    last = size;

  unsigned long f = first - 1, l = last - 1;
  T swap, element = start[(f + l) / 2];

  do {
    while (compare(start[f], element, d))
      f++;
    while (compare(element, start[l], d))
      l--;
    if (f < l && l != -1lu) {
      swap     = start[f];
      start[f] = start[l];
      start[l] = swap    ;
    }
    if (f <= l && l != -1lu) {
      f++;
      l--;
    }
  } while (f <= l && l != -1lu);
  if (first < l + 1)
    sort(d, first, l + 1);
  if (f + 1 < last)
    sort(d, f + 1, last);

  return *this;
}
template<class T> bool compare(const T &a, const T &b, const char &d) {
  return d == 'd' ? a > b : a < b;
}



/*******/
/* set */
/*******/

// general constructor:

template<class T> set<T>::set(const list<T> &l) {
  start = new T[size = l.length()];
  for (unsigned long int i = 0; i < size; i++)
    start[i] = l[i + 1];
}

// copy constructor:

template<class T> set<T>::set(const set<T> &m) {
  start = new T[size = m.size];
  for (unsigned long int i = 0; i < size; i++)
    start[i] = m.start[i];
}

// operators:

// union, intersection and minus operators

template<class T> set<T> operator+(const set<T> &m1, const set<T> &m2) {
  set<T> temp = m1;
  return temp += m2;
}

template<class T> set<T> operator^(const set<T> &m1, const set<T> &m2) {
  set<T> temp = m1;
  return temp ^= m2;
}

template<class T> set<T> operator-(const set<T> &m1, const set<T> &m2) {
  set<T> temp = m1;
  return temp -= m2;
}

template<class T> set<T>& set<T>::operator+=(const set<T> &m) {
  T add[m.size];
  unsigned long number = 0, i = 0;
  for (; i < m.size; i++)
    if (!contains(m.start[i])) {
      add[number] = m.start[i];
      number++;
    }
  if (number > 0) {
    set<T> temp = *this;
    delete[] start;
    start = new T[size += number];
    for(i = 0; i < temp.size; i++)
      start[i] = temp.start[i];
    for(i = 0; i < number; i++)
      start[i + temp.size] = add[i];
  }
  return *this;
}

template<class T> set<T>& set<T>::operator^=(const set<T> &m) {
  if (!(*this <= m)) {
    T intersect[max(size, m.size)];
    unsigned long number = 0, i = 0;
    for (; i < size; i++)
      if (m.contains(start[i])) {
        intersect[number] = start[i];
        number++;
      }
    delete[] start;
    start = new T[size = number];
    for (i = 0; i < size; i++)
      start[i] = intersect[i];
  }
  return *this;
}

template<class T> set<T>& set<T>::operator-=(const set<T> &m) {
  T newset[size];
  unsigned long number = 0, i = 0;
  for (; i < size; i++)
    if (!m.contains(start[i])) {
      newset[number] = start[i];
      number++;
    }
  if (number != size) {
    delete[] start;
    start = new T[size = number];
    for (i = 0; i < size; i++)
      start[i] = newset[i];
  }
  return *this;
}

// comparison operators:

template<class T1, class T2> bool operator==(const set<T1> &m1, const set<T2> &m2) {
  unsigned long l = m1.length();
  if (l != m2.length())
    return false;
  bool found;
  for (unsigned long i = 1; i <= l; i++) {
    found = false;
    for (unsigned long j = 1; j <= l; j++)
      if (m1[i] == m2[j]) {
        found = true;
        break;
      }
    if (!found)
      return false;
  }
  return true;
}

template<class T1, class T2> bool operator!=(const set<T1> &m1, const set<T2> &m2) {
  return !(m1 == m2);
}

template<class T1, class T2> bool operator<(const set<T1> &m1, const set<T2> &m2) {
  unsigned long l1 = m1.length(), l2 = m2.length();
  if (l1 >= l2)
    return false;
  bool found;
  for (unsigned long i = 1; i <= l1; i++) {
    found = false;
    for (unsigned long j = 1; j <= l2; j++)
      if (m1[i] == m2[j]) {
        found = true;
        break;
      }
    if (!found)
      return false;
  }
  return true;
}

template<class T1, class T2> bool operator<=(const set<T1> &m1, const set<T2> &m2) {
  unsigned long l1 = m1.length(), l2 = m2.length();
  if (l1 > l2)
    return false;
  bool found;
  for (unsigned long i = 1; i <= l1; i++) {
    found = false;
    for (unsigned long j = 1; j <= l2; j++)
      if (m1[i] == m2[j]) {
        found = true;
        break;
      }
    if (!found)
      return false;
  }
  return true;
}

template<class T1, class T2> bool operator>(const set<T1> &m1, const set<T2> &m2) {
  return m2 < m1;
}

template<class T1, class T2> bool operator>=(const set<T1> &m1, const set<T2> &m2) {
  return m2 <= m1;
}

// assignment operator:

template<class T> set<T>& set<T>::operator=(const set<T> &m) {
  delete[] start;
  start = new T[size = m.length()];
  for (unsigned long i = 0; i < size; i++)
    start[i] = m[i + 1];
}

// output operator:

template<class T> ostream& operator<<(ostream &out, const set<T> &m) {
  unsigned long len = m.length();
  if (len == 0)
    out << "{}";
  else {
    out << '{' << m[1];
    for (unsigned long i = 2; i <= len; i++)
      out << ", " << m[i];
    out << '}';
  }
  return out;
}

// methods:

template<class T> set<T>& set<T>::append(const T &a) {
  if (!contains(a))
    list<T>::insert(a, size + 1);
  return *this;
}

template<class T> bool set<T>::contains(const T &a) const {
  for (unsigned long i = 0; i < size; i++)
    if (a == start[i])
      return true;
  return false;
}

template<class T> set<T>& set<T>::insert(const T &a, unsigned long n) {
  if (contains(a) == 0)
    list<T>::insert(a, n);
  return *this;
}

template<class T> set<T>& set<T>::sort(const char &d, unsigned long first, unsigned long last) {
  *this = list<T>::sort(d, first, last);
  return *this;
}



/**********/
/* vector */
/**********/

// operators:

// addition and substraction operators:

template<class T> vector<T> operator+(const vector<T> &v) {
  return v;
}

template<class T> vector<T> operator-(const vector<T> &v) {
  unsigned long size = v.length();
  vector<T> temp = v;
  for (unsigned long i = 1; i <= size; i++)
    temp[i] = -temp[i];
  return temp;
}

template<class T> vector<T> operator+(const vector<T> &v, const vector<T> &w) {
  vector<T> temp = v;
  temp += w;
  return temp;
}

template<class T> vector<T> operator-(const vector<T> &v, const vector<T> &w) {
  vector<T> temp = v;
  temp -= w;
  return temp;
}

template<class T> vector<T>& vector<T>::operator+=(const vector<T> &v) {
  if (size != v.size)
    Error(VecSizeDiff, "vector+=vector");
  T zero;
  for (unsigned long i = 0; i < size; i++)
    if (v.start[i] != zero)
      start[i] += v.start[i];
  return *this;
}

template<class T> vector<T>& vector<T>::operator-=(const vector<T> &v) {
  *this += -v;
  return *this;
}

// multiplication and division operators:

template<class T> T operator*(const vector<T> &v, const vector<T> &w) {
  unsigned int l = v.length();
  if (l != w.length())
    Error(VecSizeDiff, "vector*vector");
  T sum = 0;
  for (unsigned long i = 1; i <= l; i++)
    sum += v[i] * w[i];
  return sum;
}

template<class T> vector<T> operator*(const vector<T> &v, const T &a) {
  vector<T> temp = v;
  temp *= a;
  return temp;
}

template<class T> vector<T> operator*(const T &a, const vector<T> &v) {
  vector<T> temp = v;
  temp *= a;
  return temp;
}

template<class T> vector<T> operator/(const vector<T> &v, const T &a) {
  vector<T> temp = v;
  temp /= a;
  return temp;
}

template<class T> vector<T>& vector<T>::operator*=(const T &a) {
  for (unsigned long i = 0; i < size; i++)
    start[i] *= a;
  return *this;
}

template<class T> vector<T>& vector<T>::operator/=(const T &a) {
  for (unsigned long i = 0; i < size; i++)
    start[i] /= a;
  return *this;
}


// comparison operators:

template<class T> bool operator==(const vector<T> &v, const vector<T> &w) {
  unsigned long l = v.length();
  if (l != w.length())
    Error(VecSizeDiff, "vector==vector");
  for (unsigned long i = 1; i <= l; i++)
    if (v[i] != w[i])
      return false;
  return true;
}

template<class T> bool operator!=(const vector<T> &v, const vector<T> &w) {
  return !(v == w);
}

// index operators:

template<class T> T& vector<T>::operator[](const unsigned long &i) {
  if (i <= 0 || i > size)
    Error(IndexRange, "vector[]");
  return start[i - 1];
}

template<class T> const T& vector<T>::operator[](const unsigned long &i) const {
  if (i <= 0 || i > size)
    Error(IndexRange, "vector[]");
  return start[i - 1];
}

// assignment operator:

template<class T> vector<T>& vector<T>::operator=(const vector<T> &v) {
  if (this != &v) {
    delete[] start;
    start = new T[size = v.size];
    for (unsigned int i = 0; i < size; i++)
      start[i] = v.start[i];
  }
  return *this;
}

// output operator:

template<class T> ostream& operator<<(ostream &out, const vector<T> &v) {
  unsigned long len = v.length();
  if (len == 0)
    out << "()";
  else {
    out << '(' << v[1];
    for (unsigned long i = 2; i <= len; i++)
      out << ',' << v[i];
    out << ')';
  }
  return out;
}

// methods:

template<class T> bool vector<T>::isnull() const {
  T zero = 0;
  for (unsigned long i = 0; i < size; i++)
    if (start[i] != zero)
      return false;
  return true;
}



/**********/
/* matrix */
/**********/

// operators:

// addition and substraction operators:

template<class T> matrix<T> operator+(const matrix<T> &M) {
  return M;
}

template<class T> matrix<T> operator-(const matrix<T> &M) {
  unsigned long lines = M.lines();
  matrix<T> temp = M;
  for (unsigned long i = 1; i <= lines; i++)
    temp[i] = -temp[i];
  return temp;
}

template<class T> matrix<T> operator+(const matrix<T> &M, const matrix<T> &N) {
  matrix<T> temp = M;
  temp += N;
  return temp;
}

template<class T> matrix<T> operator-(const matrix<T> &M, const matrix<T> &N) {
  matrix<T> temp = M;
  temp -= N;
  return temp;
}

template<class T> matrix<T>& matrix<T>::operator+=(const matrix<T> &M) {
  if (height != M.height || width != M.width)
    Error(MatSizeDiff, "matrix+=matrix");
  Lines += M.Lines;
  return *this;
}

template<class T> matrix<T>& matrix<T>::operator-=(const matrix<T> &M) {
  return *this += -M;
}

// multiplication and division operators

template<class T> matrix<T> operator*(const matrix<T> &M, const matrix<T> &N) {
  matrix<T> temp = M;
  temp *= N;
  return temp;
}

template <class T> vector<T> operator*(const matrix<T> &M, const vector<T> &v) {
  unsigned long m = M.lines(), n = M.columns();
  if (n != v.length())
    Error(MatSizeDiff, "matrix*vector");
  vector<T> temp(m);
  T sum;
  for (unsigned long i = 1; i <= m; i++) {
    sum = T(0);
    for (unsigned long j = 1; j <= n; j++)
      sum += M[i][j] * v[j];
    temp[i] = sum;
  }
  return temp;
}

template<class T> vector<T> operator*(const vector<T> &v, const matrix<T> &M) {
  unsigned long m = M.lines(), n = M.columns();
  if (m != v.length())
    Error(MatSizeDiff, "vector*matrix");
  vector<T> temp(n);
  T sum;
  for (unsigned long i = 1; i <= n; i++) {
    sum = T(0);
    for (unsigned long j = 1; j <= m; j++)
      sum += v[j] * M[j][i];
    temp[i] = sum;
  }
  return temp;
}

template<class T> matrix<T> operator*(const matrix<T> &M, const T &a) {
  matrix<T> temp = M;
  temp *= a;
  return temp;
}

template<class T> matrix<T> operator*(const T &a, const matrix<T> &M) {
  matrix<T> temp = M;
  temp *= a;
  return temp;
}

template<class T> matrix<T> operator/(const matrix<T> &M, const T &a) {
  matrix<T> temp = M;
  temp /= a;
  return temp;
}

template<class T> matrix<T>& matrix<T>::operator*=(const T &a) {
  for (unsigned i = 1; i <= height; i++)
    Lines[i] *= a;
  return *this;
}

template<class T> matrix<T>& matrix<T>::operator*=(const matrix<T> &M) {
  if (width != M.height)
    Error(MatSizeDiff, "matrix*=matrix");
  matrix<T> temp(height, M.width);
  temp.outstyle = outstyle;
  T sum;
  for (unsigned long i = 1; i <= height; i++)
    for (unsigned long j = 1; j <= width; j++) {
      sum = T(0);
      for (unsigned long k = 1; k <= width; k++)
        sum += Lines[i][k] * M[k][j];
      temp[i][j] = sum;
    }
  *this = temp;
  return *this;
}

template<class T> matrix<T>& matrix<T>::operator/=(const T &a) {
  for (unsigned i = 1; i <= height; i++)
    Lines[i] /= a;
  return *this;
}

template<class T> matrix<T>& matrix<T>::operator^=(const Z &a) {
  Z expo = a;
  if (height != width)
    Error(MatNotSquare, "matrix^=Z");
  if (a < 0) {
    invert();
    expo = -a;
  }
  matrix<T> mult = *this, prod(height, "1");
  prod.outstyle = outstyle;
  for (;expo != 0; expo /= 2) {
    if (expo % 2 == 1)
      prod *= mult;
    if (expo > 0)
      mult *= mult;
  }
  *this = prod;
  return *this;
}


// comparison operators:

template<class T> bool operator==(const matrix<T> &M, const matrix<T> &N) {
  unsigned long m = M.lines(), n = M.columns();
  if (m != N.lines() || n != N.columns())
    return false;
  for (unsigned long i = 1; i <= m; i++)
    for (unsigned long j = 1; j <= n; j++)
    if (M[i][j] != N[i][j])
      return false;
  return true;
}

template<class T> bool operator!=(const matrix<T> &M, const matrix<T> &N) {
  return !(M == N);
}

// index operators:

template<class T> vector<T>& matrix<T>::operator[](const unsigned long &i) {
  return Lines[i];
}

template<class T> const vector<T>& matrix<T>::operator[](const unsigned long &i) const {
  return Lines[i];
}

// assignment operators:

template<class T> matrix<T>& matrix<T>::operator=(const matrix<T> &M) {
  if (*this != M) {
    height   = M.height  ;
    width    = M.width   ;
    outstyle = M.outstyle;
    Lines    = M.Lines   ;
  }
  return *this;
}

template<class T> matrix<T>& matrix<T>::operator=(const vector<T> &v) {
  height = v.length();
  width = 1;
  Lines = vector<Line>(height, vector<T>(1));
  for (unsigned long i = 1; i <= height; i++)
    Lines[i][1] = v[i];
  return *this;
}

// output operator:

template<class T> ostream& operator<<(ostream &out, const matrix<T> &M) {
  unsigned long i, j, m = M.lines(), n = M.columns();
  string style = M.getoutstyle();
  if (style == "standard") {
    out << '[';
    for (i = 1; i <= m; i++) {
      for (j = 1; j <= n; j++) {
        out << M[i][j];
        if (j < n)
          out << ", ";
      }
      if (i < m)
        out << "; ";
    }
    out << ']';
    return out;
  }
  if (style == "lines") {
    out << endl << '[';
    for (i = 1; i <= m; i++) {
      if (i > 1)
        out << endl << ' ';
      for (j = 1; j <= n; j++) {
        out << M[i][j];
        if (j < n)
          out << ", ";
      }
      if (i < m)
        out << "; ";
    }
    out << ']' << endl;
    return out;
  }
  if (style == "matrix") {
    list<unsigned long> widths(n, 0);
    unsigned long k, colwidth, startcolumn = 1, endcolumn = 1, lspace, rspace;
    string s;
    for (j = 1; j <= n; j++) {
      colwidth = 0;
      for (i = 1; i <= m; i++) {
        colwidth = max(colwidth, string(M[i][j]).size());
      }
      widths[j] = colwidth;
    }
    colwidth = 2;
    out << endl;
    while (endcolumn <= n) {
      colwidth += widths[endcolumn];
      while (endcolumn < n && colwidth + widths[endcolumn + 1] + 1 + (endcolumn + 1 == n ? 2 : 0) <= 80) {
        endcolumn++;
        colwidth += widths[endcolumn] + 1;
      }
      if (startcolumn == 1)
        out << (unsigned char) 218;
      else
        out << ' ';
      if (endcolumn == n)
        out << space(colwidth) << (unsigned char) 191;
      if (endcolumn < n || colwidth + (startcolumn == 1 ? 2 : 0) < 78)
        out << endl;
      for (i = 1; i <= m; i++)
        for (j = startcolumn; j <= endcolumn; j++) {
          if (j == 1)
            out << (unsigned char) 179 << ' ';
          s = string(M[i][j]);
          lspace = widths[j] - s.size();
          rspace = lspace / 2;
          lspace -= rspace;
          out << space(lspace) << s << space(rspace);
          if (j == endcolumn) {
            if (j == n)
              out << ' ' << (unsigned char) 179;
            if (colwidth + (j == n ? 2 : 0)< 80)
              out << endl;
          }
          else
            out << ' ';
        }
      if (startcolumn == 1)
        out << (unsigned char) 192;
      else
        out << ' ';
      if (endcolumn == n)
        out << space(colwidth) << (unsigned char) 217;
      out << endl;
      startcolumn = ++endcolumn;
      colwidth = 0;
    }
    out << endl;
    return out;
  }
}


// methods:

template<class T> unsigned long matrix<T>::columns() const {
  return width;
}

/*template<class T> T matrix<T>::det() const {
cout << endl << "Matrix: " << height << " x " << width << endl;
  if (height != width)
    Error(MatNotSquare, "matrix::det()");
  if (height == 1)
    return Lines[1][1];
  matrix<T> temp = *this;
  unsigned long pos = 0;
  T mini = 0, l, number, div = 1, fact;
  string ct = mini.classtype();
  for (unsigned long i = 1; i <= height; i++)
    if (Lines[i][1] != T(0) && (mini == T(0) || Lines[i][1] < mini)) {
      pos = i;
      mini = Lines[i][1];
    }
  if (pos == 0)
    return 0;
  if (ct == "ring") {
    for (unsigned long i = 1; i <= height; i++) {
      number = Lines[i][1];
      if (number != 0 && i != pos) {
        l = lcm(mini, number);
        fact = l / number;
        div *= fact;
        temp.Lines[i] = fact * Lines[i] - l / mini * Lines[pos];
      }
    }
    return mini * temp.leaveout(pos, 1).det() / div;
  }
}*/
/*template<class T> T matrix<T>::det() const {
  if (height != width)
    Error(MatNotSquare, "matrix::det()");
  if (height == 1)
    return Lines[1][1];
  // develop for first column
  T sum = 0, sign = 1;
  for (unsigned long i = 1; i <= height; i++, sign = -sign)
    sum += Lines[i][1] * sign * leaveout(i, 1).det();
  return sum;
}*/

template<class T> string matrix<T>::getoutstyle() const {
  return outstyle;
}

/*template<class T> matrix<T> matrix<T>::leaveout(unsigned long p, unsigned long q) const {
cout << endl << "Leave out: Matrixdimension: " << height << " x " << width << endl;
  unsigned long m, n;
  if (p == 0) {
    m = height;
    p = m + 1;
  }
  else
    m = height - 1;
  if (q == 0) {
    n = width;
    q = n + 1;
  }
  else
    n = width - 1;
cout << "m = " << m << "  n = " << n << endl;
  matrix<T> temp(m, n);
cout << "m = " << m << "  n = " << n << endl;
  temp.outstyle = outstyle;
  for (unsigned long i = 1; i <= m; i++)
    for (unsigned long j = 1; j <= n; j++)
      temp[i][j] = Lines[i < p ? i : i + 1][j < q ? j : j + 1];
  return temp;
}*/

template<class T> unsigned long matrix<T>::lines() const {
  return height;
}

/*template<class T> matrix<T> matrix<T>::minor(unsigned long m, unsigned long n, unsigned long h, unsigned long w) const {
  if (m > height || n > width)
    return matrix<T>(0, 0);
  unsigned long mdiff = m - 1                                     ,
                ndiff = n - 1                                     ,
                mlast = (h == 0 ? height : min(height, mdiff + h)),
                nlast = (w == 0 ? width  : min(width , ndiff + w));
  h = mlast - mdiff;
  w = nlast - ndiff;
  matrix<T> temp(h, w);
  temp.outstyle = outstyle;
  for (unsigned long i = 1; i <= temp.height; i++)
    for (unsigned long j = 1; j <= temp.width; j++)
      temp.Lines[i][j] = Lines[i + m - 1][j + n - 1];
  return temp;
}*/

template<class T> void matrix<T>::setoutstyle(const string &s) {
  if (s == "standard" || s == "lines" || s == "matrix")
    outstyle = s;
}

template<class T> matrix<T> matrix<T>::t() const {
  matrix<T> temp(width, height);
  temp.outstyle = outstyle;
  for (unsigned long i = 1; i <= width; i++)
    for (unsigned long j = 1; j <= height; j++)
      temp[i][j] = Lines[j][i];
  return temp;
}

template<class T> matrix<T> matrix<T>::transpose() {
  *this = t();
  return *this;
}


#endif

