#include <iostream>
#include <string>
#include "error.h"

void Error_Execute(ErrorCode &code, string &source, string &message, ErrorLevel &level) {
  switch (level) {
    case ignore: break;
    case warn:   cout << endl << "Warning! Error (code " << code << ") in function " << source << ": " << message << endl;
                 break;
    default:     cerr << "Error (code " << code << ") in function " << source << ": " << message << endl;
                 exit(-1);
  }
}

void Error(ErrorCode code, string source, ErrorLevel level) {
  string     message;
  switch (code) {
    case DivByZero     : message = "Division by 0!"                         ; break;
    case ExpLessZero   : message = "Exponent must not be less than 0!"      ; break;
    case IndetExists   : message = "Indeterminate already exists!"          ; break;
    case IndetNoLetter : message = "Indeterminate must be a letter!"        ; break;
    case IndetNotExist : message = "Indeterminate does not exist!"          ; break;
    case IndetNotValid : message = "Not a valid name for an indeterminate!" ; break;
    case IndexRange    : message = "Index is not in range!"                 ; break;
    case MatIdNotOne   : message = "Identity character has to be \"1\"!"    ; break;
    case MatNotSquare  : message = "Matrix is not a square matrix"          ; break;
    case MatSizeDiff   : message = "Matrix sizes do not match"              ; break;
    case MonoLogLen    : message = "Size of log list does not match!"       ; break;
    case MonoNotValid  : message = "Not a valid monomial!"                  ; break;
    case NoPrime       : message = "Need a primary number!"                 ; break;
    case NotDiv        : message = "Cannot divide objects!"                 ; break;
    case PolyDiffDeg   : message = "Polynomials must have same the degree!" ; break;
    case PolyManyIndet : message = "There are not so many monomials!"       ; break;
    case PolyNotHomo   : message = "Polynomial is not homogenous!"          ; break;
    case TermNoAdd     : message = "Terms cannot be summed!"                ; break;
    case VecSizeDiff   : message = "Vector sizes do not match!"             ; break;
    case MatVecSizeDiff: message = "Row length does not match matrix width!"; break;
    default            : message = "Unknown Error!"                         ;
  }
  Error_Execute(code, source, message, level);
}

