#include "contain.t"
#include "mathfunc.h"
#include "mathops.h"
#include "z.h"

list<Z> ExtEuklid(const Z &a, const Z &b) {
  Z Q;
  list<Z> L1(3, 0), L2(3, 0), L3(3, 0);

  Q     = abs(a);
  L1[1] = Q / a ;
  L1[3] = Q     ;
  Q     = abs(b);
  L2[2] = Q / b ;
  L2[3] = Q     ;

  if (a == 0)
    if (b == 0)
      return L3;
    else
      return L2;
  else
    if (b == 0)
      return L1;

  while (L2[3] != 0) {
    Q = L1[3] / L2[3];
    L3[1] = L1[1] - Q * L2[1];
    L3[2] = L1[2] - Q * L2[2];
    L3[3] = L1[3] - Q * L2[3];
    L1 = L2;
    L2 = L3;
  }
  return L1;
}

bool IsPrime(const Z &p) {
  static Z lastp = 0;
  static bool lastreturned = false;
  if (p == lastp)
    return lastreturned;
  lastp = p;
  lastreturned = RabinMiller(p);
  return lastreturned;
}

Z ModPot(Z a, Z n, const Z &m) {
  Z p = 1;
  while (n > 0) {
    if (isodd(n))
      p = p * a % m;
    a = (a^2) % m;
    n /= 2;
  }
  return p;
}

bool RabinMiller(const Z &n) {
  if (n == 2)
    return true;
  if (n < 2 || iseven(n))
    return false;

  Z b = Z(4)^10, n1 = n - 1, s = 0, t = n1;
  unsigned long i, j, k = 10;
  while (b <= n) {
    k++;
    b *= 4;
  }
  while (iseven(t)) {
    s++;
    t /= 2;
  }

  for (i = 1; i <= k; i++) {
    b = Rand(1, n1);
    if (gcd(b, n) > 1)
      return false;
    b = ModPot(b, t, n);
    if (b != 1 && b != n1)
      for (j = 1; j < s; j++) {
        b = ModPot(b, 2, n);
        if (b == n1)
          break;
        if (b == 1)
          return false;
      }
  }
  return true;
}

