#ifndef mathops_h
#define mathops_h

#include <cmath>
#include "z.h"
#include "contain.t"

// function declarations:

                        Z       gcd      (      Z       ,       Z       );
                        Z       lcm      (const Z      &, const Z      &);
template<class numeric> numeric max      (const numeric&, const numeric&);
template<class numeric> numeric min      (const numeric&, const numeric&);
template<class numeric> numeric binomial (const numeric&, const numeric&);


// template implementation:

template<class numeric> numeric max(const numeric &a, const numeric &b) {
  return a > b ? a : b;
}

template<class numeric> numeric min(const numeric &a, const numeric &b) {
  return a < b ? a : b;
}

template<class numeric> numeric binomial(const numeric &n, const numeric &k) {
  numeric zero(0),
          one (1);
  static list<list<numeric> > PascalTriangle;

  // checking for special inputs
  if (k < 0)
    if (n < k || n >= zero)
      return zero;
    else
      return binomial(n, n - k);
  if (n < 0)
    return (k % numeric(2) == zero ? one : -one) * binomial(k - n - one, k);
  if (k > n)
    return zero;

  // calculating binomial coefficient
  for (unsigned long i = PascalTriangle.length() + 1; i <= n + 1; i++) {
    PascalTriangle.append(list<numeric>(i));
    PascalTriangle[i][1] = 1;
    PascalTriangle[i][i] = 1;
    for (unsigned long j = 2; j < i; j++)
      PascalTriangle[i][j] = PascalTriangle[i - 1][j - 1] + PascalTriangle[i - 1][j];
  }

  return PascalTriangle[n + 1][k + 1];
}

#endif

