#include <iostream>
#include <string>
#include "error.h"
#include "poly.h"
#include "funcs.h"
#include "mathops.h"
#include "z.h"
#include "contain.t"

bool debugpoly = false;

/******************/
/* indeterminates */
/******************/

indet_list              IndetList;
Indet_Ops               indets   ;
list<list<vector<Z> > > MonoTable;

// methods:

void indet_list::add(indet &x) {
  unsigned long addcount = x.index ? x.end - x.begin + 1 : 1,
                addindex = x.index ? x.begin : -1lu         ,
                delcount = 0                                ,
                pos      = 0                                ;
  for (; pos < entries && name[pos] < x.name; pos++);
  unsigned long i = pos, lastadd = pos + addcount - 1;
  for (; i < entries && x.name == name[i]; i++, delcount++);
  unsigned long firstafter = i;
  unsigned long  tempentries = entries;
  char         * tempname    = name   ;
  unsigned long* tempindex   = index  ;
  indet*       * temppoint   = point  ;
  entries += addcount - delcount;
  name  = new char         [entries];
  index = new unsigned long[entries];
  point = new indet*       [entries];
  for (i = 0; i < pos; i++) {
    name [i] = tempname [i];
    index[i] = tempindex[i];
    point[i] = temppoint[i];
  }
  for (; i <= lastadd; i++, addindex++) {
    name [i] = x.name  ;
    index[i] = addindex;
    point[i] = &x      ;
  }
  for (; firstafter < tempentries; firstafter++, i++) {
    name [i] = tempname [firstafter];
    index[i] = tempindex[firstafter];
    point[i] = temppoint[firstafter];
  }
  delete[] tempname ;
  delete[] tempindex;
  delete[] temppoint;
}

void indet_list::clear() {
  delete[] name ;
  delete[] index;
  delete[] point;
  entries =                   0 ;
  name    = new char         [0];
  index   = new unsigned long[0];
  point   = new indet*       [0];
}

void indet_list::remove(const indet &x) {
  unsigned long pos = search(x.name) - 1, remcount = 0;
  if (pos != -1lu) {
    unsigned long i = pos;
    char         * tempname    = name ;
    unsigned long* tempindex   = index;
    indet*       * temppoint   = point;
    for (; i < entries && name[i] == x.name; remcount++, i++);
    entries -= remcount;
    name  = new char         [entries];
    index = new unsigned long[entries];
    point = new indet*       [entries];
    for (i = 0; i < pos; i++) {
      name [i] = tempname [i];
      index[i] = tempindex[i];
      point[i] = temppoint[i];
    }
    for (; i < entries; i++) {
      name [i] = tempname [i + remcount];
      index[i] = tempindex[i + remcount];
      point[i] = temppoint[i + remcount];
    }
    delete[] tempname ;
    delete[] tempindex;
    delete[] temppoint;
  }
}

unsigned long indet_list::search(const char &indetname) {
  for (unsigned long i = 0; i < entries; i++)
    if (name[i] == indetname)
      return i + 1;
  return 0;
}

unsigned long indet_list::search(const char &indetname, const unsigned long &idx) {
  unsigned long i = search(indetname);
  if (i)
    for (i--; i < entries; i++) {
      if (name[i] != indetname)
        return 0;
      if (index[i] == idx)
        return i + 1;
    }
  return 0;
}

unsigned long indet_list::searchex(const char &indetname) {
  unsigned long i = search(indetname);
  if (i && index[i - 1] == -1lu)
    return i;
  return 0;
}


/**************************/
/* indeterminate handling */
/**************************/

// destructor:

Indet_Ops::~Indet_Ops() {
  Indet_Listing* del;
  while (first != NULL) {
    del   = first      ;
    first = first->next;
    delete del;
  }
}

// methods:

void Indet_Ops::add(const char &indetname) {
  add(indetname, 1, 0);
}

void Indet_Ops::add(const char &indetname, const unsigned long &n) {
  add(indetname, n, n);
}

void Indet_Ops::add(const char &indetname, const unsigned long &m, const unsigned long &n) {
  if (indetname < 'A' || indetname > 'z' || indetname > 'Z' && indetname < 'a')
    Error(IndetNoLetter, "Indet_Ops::add()");
  if (IndetList.search(indetname))
    Error(IndetExists, "Indet_Ops::add()");

  if (first == NULL) {
    first = new Indet_Listing;
    last  = first            ;
  }
  else {
    last->next = new Indet_Listing;
    last       = last->next       ;
  }
  last->ind.name = indetname;
  if (m > n)
    last->ind.reindex();
  else
    last->ind.reindex(m, n);
  MonoTable.clear();
}

void Indet_Ops::CalcMonoLine(const unsigned long& d) {
  if (d == 0 || MonoTable.length() >= d && MonoTable[d].length() > 0)
    return;
  unsigned long i                           ,
                n   = IndetList.entries     ,
                l   = binomial(n + d - 1, d),
                pos = 1                     ;
  vector<Z> L(n);
  if (MonoTable.length() < d)
    MonoTable.append(list<vector<Z> >(0), d - MonoTable.length());
  MonoTable[d] = list<vector<Z> >(l, L);
  L[1] = d;
  MonoTable[d][1] = L;
  switch (TermOrd) {
    case Lex      :
    case DegLex   : for (i = 2; i <= l; i++) {
                      if (pos == n) {
                        pos--;
                        while (L[pos] == 0)
                          pos--;
                      }
                      L[pos++]--;
                      L[pos  ]++;
                      if (pos != n && L[n] != 0) {
                        L[pos] += L[n];
                        L[n  ]  = 0   ;
                      }
                      MonoTable[d][i] = L;
                    }
                    return;
    case DegRevLex: for (i = 2; i <= l; i++) {
                      if (L[1] == 0) {
                        pos = 2;
                        while (L[pos] == 0)
                          pos++;
                        L[pos + 1]++;
                        L[1  ] = L[pos] - 1;
                        L[pos] = 0         ;
                      }
                      else {
                        L[1]--;
                        L[2]++;
                      }
                      MonoTable[d][i] = L;
                    }
                    return;
    case Elim     : for (i = 2; i <= l; i++) {
                      if (L[1] == 0) {
                        pos = 2;
                        while (L[pos] == 0 && pos < TermOrdElim)
                          pos++;
                        if (pos == TermOrdElim) {
                          pos++;
                          while (L[pos] == 0)
                            pos++;
                          if (pos >= n) {
                            L[1] = L[TermOrdElim] - 1;
                            if (TermOrdElim != 1)
                              L[TermOrdElim] = 0;
                            L[TermOrdElim + 1] = L[n] + 1;
                            if (n != TermOrdElim + 1)
                              L[n] = 0;
                          }
                          else {
                            L[1] = L[TermOrdElim];
                            if (TermOrdElim != 1)
                              L[TermOrdElim] = 0;
                            L[pos + 1]++;
                            L[TermOrdElim + 1] = L[pos] - 1;
                            if (pos != TermOrdElim + 1)
                              L[pos] = 0;
                          }
                        }
                        else {
                          L[pos + 1]++;
                          L[1  ] = L[pos] - 1;
                          L[pos] = 0         ;
                        }
                      }
                      else {
                        L[1]--;
                        L[2]++;
                      }
                      MonoTable[d][i] = L;
                    }
                    return;
  }
}

monomial Indet_Ops::GetHomoMono(const unsigned long &d, const unsigned long &position) {
  if (d == 0)
    return monomial();
  list<Z> L;
  CalcMonoLine(d);
  L = MonoTable[d][position];
  return LogToMono(L);
}

unsigned long Indet_Ops::GetHomoPos(const monomial &m) {
  if (m == monomial())
    return 1;
  unsigned long pos = 1      ,
                d   = m.deg();
  vector<Z>     lg  = m.log();
  CalcMonoLine(d);
  while (MonoTable[d][pos] != lg)
    pos++;
  return pos;

}

string Indet_Ops::GetList() {
  bool found = false;
  char lastname = 0;
  string indetlist;
  for (unsigned long i = 0; i < IndetList.entries; i++)
    if (IndetList.name[i] != lastname) {
      lastname = IndetList.name[i];
      if (found)
        indetlist += ", ";
      else
        found = true;
      indetlist += lastname;
      if (IndetList.point[i]->index) {
        indetlist += '[';
        indetlist += str(IndetList.point[i]->begin);
        if (IndetList.point[i]->begin < IndetList.point[i]->end) {
          indetlist += "..";
          indetlist += str(IndetList.point[i]->end);
        }
        indetlist += ']';
      }
    }
  return indetlist;
}

void Indet_Ops::modify(const char &indetname) {
  unsigned long pos = IndetList.search(indetname);
  if (pos)
    IndetList.point[pos]->reindex();
  else
    add(indetname, 1, 0);
}

void Indet_Ops::modify(const char &indetname, const unsigned long &n) {
  unsigned long pos = IndetList.search(indetname);
  if (pos)
    IndetList.point[pos - 1]->reindex(n, n);
  else
    add(indetname, n, n);
}

void Indet_Ops::modify(const char &indetname, const unsigned long &m, const unsigned long &n) {
  if (m > n)
    modify(indetname);
  else {
    unsigned long pos = IndetList.search(indetname);
    if (pos)
      IndetList.point[pos - 1]->reindex(m, n);
    else
      add(indetname, m, n);
  }
}

void Indet_Ops::remove(const char &indetname) {
  Indet_Listing *scan = first, *lastscan = NULL;
  while (scan != NULL) {
    if (scan->ind.name == indetname) {
      IndetList.remove(scan->ind);
      if (scan->next == NULL) {
        last = lastscan;
        if (last == NULL)
          first = NULL;
        else
          last->next = NULL;
      }
      else
        if (scan == first)
          first = scan->next;
        else
          lastscan->next = scan->next;
      delete scan;
      return;
    }
    else {
      lastscan = scan;
      scan = scan->next;
    }
  }
}

void Indet_Ops::removeall() {
  Indet_Listing* del;
  while (first != NULL) {
    del   = first      ;
    first = first->next;
    delete del;
  }
  last = NULL;
  IndetList.clear();
}

/******************/
/* power products */
/******************/

// general constructor:

pp::pp(const string &s, const Z &a = 1) {
  char c = s[0];
  if (isalpha(c)) {
    unsigned long pos = IndetList.search(c);
    if (!pos)
      Error(IndetNotExist, "pp()");
    else {
      unsigned long l = s.length();
      ind = IndetList.point[pos - 1];
      if (l > 1) {
        if (ind->index) {
          char b1 = s[1], b2 = s[l - 1];
          if (b1 != '[' || b2 != ']')
            Error(IndetNotValid, "pp()");
          else {
            string sub = s.substr(2, l - 3);
            index = val(sub);
            if (index < ind->begin || index > ind->end)
              Error(IndexRange, "pp()");
            else {
              Indet    = s;
              Exponent = a;
            }
          }
        }
        else
          Error(IndetNotExist, "pp()");
      }
      else {
        Indet    =  c  ;
        Exponent =  a  ;
        index    = -1lu;
      }
    }
  }
  else
    Error(IndetNotExist, "pp()");
}

// operators:

// multiplication and division operators:

pp operator/(const pp &x, const pp &y) {
  pp temp = x;
  temp /= y;
  return temp;
}

pp operator/(const pp &x, const monomial &m) {
  pp temp = x;
  temp /= m;
  return temp;
}

pp operator^(const pp &x, const Z &a) {
  pp temp = x;
  temp ^= a;
  return temp;
}

pp pp::operator/=(const pp &x) {
  if (Indet != x.Indet && x.Exponent > 0 || Exponent < x.Exponent)
    Error(NotDiv, "pp/=pp");
  Exponent -= x.Exponent;
  return *this;
}

pp pp::operator/=(const monomial &m) {
  if (m.ppcount > 1)
    Error(NotDiv, "pp/=monomial");
  if (m.ppcount == 1)
    if (m.pps[1].Indet != Indet || m.pps[1].Exponent > Exponent)
      Error(NotDiv, "pp/=monomial");
    else
      Exponent -= m.pps[1].Exponent;
  return *this;
}

pp pp::operator^=(const Z &a) {
  Exponent *= a;
  return *this;
}

// comparison operators:

bool operator==(const pp &x, const pp &y) {
  return x.ind == y.ind && x.index == y.index && x.Exponent == y.Exponent;
}

bool operator!=(const pp &x, const pp &y) {
  return !(x == y);
}

// output operator:

ostream& operator<<(ostream &out, const pp &x) {
  if (x.Exponent == 0)
    out << 1;
  else {
    out << x.Indet;
    if (x.Exponent != 1)
      out << '^' << x.Exponent;
  }
  return out;
}

// methods:

Z pp::deg() const {
  return Exponent;
}


/*************/
/* monomials */
/*************/

// general constructors:

monomial::monomial(const string &s):
  ppcount(0) {
  if (!isalpha(s[0]))
    if (s == "1")
      return;
    else
      Error(MonoNotValid, "monomial()");
  string temp;
  pp inspp;
  unsigned long first = 0, last, l = s.length(), i, exppos;
  Z exp;
  bool done;
  while (first < l) {
    while(!isalpha(s[first]) && first < l)
      first++;
    last = first + 1;
    while(!isalpha(s[last]) && last < l)
      last++;
    temp = s.substr(first, last - first);
    exppos = temp.rfind("^");
    if (exppos == -1lu)
      inspp = pp(temp);
    else
      inspp = pp(temp.substr(0, exppos), Z(temp.substr(exppos + 1)));
    done = false;
    for (i = 1; i <= ppcount; i++) {
      if (pps[i].Indet[0] > inspp.Indet[0])
        break;
      if (pps[i].Indet[0] == inspp.Indet[0]) {
        if (inspp.ind->index) {
          while(i <= ppcount && pps[i].Indet[0] == inspp.Indet[0] && pps[i].index < inspp.index)
            i++;
          if (i > ppcount || pps[i].Indet[0] != inspp.Indet[0] || pps[i].index != inspp.index)
            break;
        }
        pps[i].Exponent *= inspp.Exponent;
        done = true;
        break;
      }
    }
    if (!done) {
      pps.insert(inspp, i);
      ppcount++;
    }
    first = last;
  }
}

monomial::monomial(const pp &x):
  ppcount(1) {
  pps.append(x);
}

// multiplication and division operators:

monomial operator* (const pp &x, const pp &y) {
  monomial temp = x;
  temp *= y;
  return temp;
}

monomial operator* (const pp &x, const monomial &m) {
  monomial temp = m;
  temp *= x;
  return temp;
}

monomial operator* (const monomial &m, const pp &x) {
  monomial temp = m;
  temp *= x;
  return temp;
}

monomial operator* (const monomial &m, const monomial &n) {
  monomial temp = m;
  temp *= n;
  return temp;
}

monomial operator/ (const monomial &m, const pp &x) {
  monomial temp = m;
  temp /= x;
  return temp;
}

monomial operator/ (const monomial &m, const monomial &n) {
  monomial temp = m;
  temp /= n;
  return temp;
}

monomial monomial::operator*=(const pp &x) {
  unsigned long i = 1;
  while (i <= ppcount && pps[i].Indet < x.Indet)
    i++;
  if (i <= ppcount && pps[i].Indet == x.Indet)
    pps[i].Exponent += x.Exponent;
  else {
    pps.insert(x, i);
    ppcount++;
  }
  return *this;
}

monomial monomial::operator*=(const monomial &m) {
  for (unsigned long i = 1; i <= m.ppcount; i++)
    *this *= m.pps[i];
  return *this;
}

monomial monomial::operator/=(const pp &x) {
  unsigned long pos = 0;
  for (unsigned long i = 1; i <= ppcount; i++)
    if (pps[i].Indet == x.Indet) {
      pos = i;
      break;
    }
  if (!pos && x.Exponent > 0 || pps[pos].Exponent < x.Exponent)
    Error(NotDiv, "monomial/=pp");
  else {
    pps[pos].Exponent -= x.Exponent;
    if (pps[pos].Exponent == 0) {
     pps.remove(pos);
      ppcount--;
    }
  }
  return *this;
}

monomial monomial::operator/=(const monomial &m) {
  list<unsigned long> pos(m.ppcount, 0);
  for (unsigned long i = 1; i <= m.ppcount; i++) {
    for (unsigned long j = 1; j <= ppcount; j++)
      if (pps[j].Indet == m.pps[i].Indet) {
        pos[i] = j;
        break;
      }
    if (!pos[i] && m.pps[i].Exponent > 0 || pps[pos[i]].Exponent < m.pps[i].Exponent)
      Error(NotDiv, "monomial/=monomial");
  }
  for (unsigned long i = 1; i <= m.ppcount; i++)
    pps[pos[i]].Exponent -= m.pps[i].Exponent;
  pos.sort('d');
  for (unsigned long i = 1; i <= m.ppcount; i++)
    if (pps[pos[i]].Exponent == 0) {
      pps.remove(pos[i]);
      ppcount--;
    }
  return *this;
}

monomial monomial::operator^=(const Z &a) {
  for (unsigned long i = 1; i <= ppcount; i++)
    pps[i] ^= a;
  return *this;
}

monomial operator^(const monomial &m, const Z &a) {
  monomial temp = m;
  temp ^= a;
  return temp;
}

// comparison operators:

bool operator==(const monomial &m, const monomial &n) {
  if (m.ppcount != n.ppcount)
    return false;
  for (unsigned long i = 1; i <= m.ppcount; i++)
    if (!n.pps.seek(m.pps[i]))
      return false;
  return true;
}

bool operator!=(const monomial &m, const monomial &n) {
  return !(m == n);
}

bool operator<(const monomial &m, const monomial &n) {
  unsigned long i;
  list<Z> logm = m.log(), logn = n.log();
  vector<Z> log = vector<Z>(logn) - vector<Z>(logm);
  Z dm, dn;

  // elimination degree:
  if (TermOrd == Elim) {
    unsigned long last = IndetList.entries - TermOrdElim;
    dm = 0;
    dn = 0;
    for (i = 1; i <= last; i++) {
      dm += logm[i];
      dn += logn[i];
    }
    if (dm < dn)
      return true;
    if (dm > dn)
      return false;
  }

  // degree compatibility:
  if (TermOrd == DegLex || TermOrd == DegRevLex || TermOrd == Elim) {
    dm = m.deg();
    dn = n.deg();
    if (dm < dn)
      return true;
    if (dm > dn)
      return false;
  }

  // lexicographical orders:
  if (TermOrd == Lex || TermOrd == DegLex) {
    for (i = 1; i <= log.length(); i++)
      if (log[i] != 0)
        if (log[i] > 0)
          return true;
        else
          return false;
    return false;
  }

  // reverse lexicographical orders:
  if (TermOrd == DegRevLex || TermOrd == Elim) {
    for (i = log.length(); i > 0; i--)
      if (log[i] != 0)
        if (log[i] < 0)
          return true;
        else
          return false;
    return false;
  }
}

bool operator>(const monomial &m, const monomial &n) {
  return n < m;
}

bool operator<=(const monomial &m, const monomial &n) {
  return m == n || m < n;
}

bool operator>=(const monomial &m, const monomial &n) {
  return m == n || n < m;
}


// assignment operators:

monomial monomial::operator=(const pp &x) {
  ppcount = 1;
  pps.clear();
  pps.append(x);
  return *this;
}

monomial monomial::operator=(const monomial &m) {
  ppcount = m.ppcount;
  pps     = m.pps    ;
  return *this;
}


// output operator:

ostream& operator<<(ostream &out, const monomial &m) {
  if (m.ppcount == 0)
    out << 1;
  else
    for (unsigned int i = 1; i <= m.ppcount; i++)
      out << m.pps[i];
  return out;
}

// methods:

Z monomial::deg() const {
  Z degree = 0;
  for (unsigned i = 1; i <= ppcount; degree += pps[i].Exponent, i++);
  return degree;
}

bool monomial::devides(const monomial &m) const {
  unsigned long i, j;
  for (i = 1; i <= ppcount; i++) {
    for (j = 1; j <= m.ppcount; j++)
      if (pps[i].Indet == m.pps[j].Indet) {
        if (pps[i].Exponent > m.pps[j].Exponent)
          return false;
        break;
      }
    if (j > m.ppcount)
      return false;
  }
  return true;
}

list<Z> monomial::log() const {
  list<Z> loglist(IndetList.entries, 0);
  for (unsigned long i = 1; i <= ppcount; i++)
    loglist[IndetList.search(pps[i].ind->name, pps[i].index)] = pps[i].Exponent;
  return loglist;
}

// functions:

monomial gcd(const monomial &m, const monomial &n) {
  monomial g;
  unsigned long i = 1, j;
  for (i = 1; i <= m.ppcount; i++)
    for (j = 1; j <= n.ppcount; j++)
      if (m.pps[i].Indet == n.pps[j].Indet) {
        g *= pp(m.pps[i].Indet, min(m.pps[i].Exponent, n.pps[j].Exponent));
        break;
      }
  return g;
}

monomial lcm(const monomial &m, const monomial &n) {
  return m * n / gcd(m, n);
}

monomial LogToMono(const list<Z> &lg) {
  monomial m;
  unsigned long l = lg.length();
  string indetname;
  if (l != IndetList.entries)
    Error(MonoLogLen, "LogToMono()");
  for (unsigned long i = 1; i <= l; i++)
    if (lg[i] != 0) {
      indetname = IndetList.name[i - 1];
      if (IndetList.point[i - 1]->index) {
        indetname += '[';
        indetname += str(IndetList.index[i - 1]);
        indetname += ']';
      }
      m *= pp(indetname, lg[i]);
    }
  return m;
}


/**********/
/* fields */
/**********/

// active field:

FIELD Field = RAT;


/***************/
/* term orders */
/***************/

// active term order:

TO            TermOrd = DegRevLex;
unsigned long TermOrdElim        ;

// function for setting active term order:

void termorder(const TO &to, const unsigned long &n) {
  if (to == Elim) {
    if (n == 0 || n >= IndetList.entries)
      TermOrd = DegRevLex;
    else {
      TermOrd     = Elim;
      TermOrdElim = n;
    }
    return;
  }
  if (to == Ord)
    return; // Ord is not implemented yet
  if (to <= DegRevLex)
    TermOrd = to;
  MonoTable.clear();
}

