#include <iostream>
#include <limits>
#include <string>
#include "error.h"
#include "mathops.h"
#include "z.h"

#include <fstream>

/***********************/
/* methods for class Z */
/***********************/

// constructors:

Z::Z(const int &n) {
  short i = sizeof(int) - 1;
  const int *ip = &n;
  unsigned int l, *lp;
  unsigned char *ic;

  ic = (unsigned char*)ip;
  if (ic[i] < 128) { // n is positive
    l = n;
    sign = 1;
  }
  else {             // n is negative
    l = -n;
    sign = -1;
  }
  lp = &l;
  ic = (unsigned char*)lp;
  while(ic[i] == 0 && --i >= 0);
  start = new unsigned char[size = i + 1];
  for(unsigned long j = 0; j < size; j++)
    start[j] = ic[j];
}

Z::Z(const unsigned int &n):
  sign(1) {
  short i = sizeof(int) - 1;
  const unsigned int *ip = &n;
  unsigned char *ic;

  ic = (unsigned char*)ip;
  while(ic[i] == 0 && --i >= 0);
  start = new unsigned char[size = i + 1];
  for(unsigned long j = 0; j < size; j++)
    start[j] = ic[j];
};

Z::Z(const long &n) {
  short i = sizeof(long) - 1;
  const long *ip = &n;
  unsigned long l, *lp;
  unsigned char *ic;

  ic = (unsigned char*)ip;
  if (ic[i] < 128) { // n is positive
    l = n;
    sign = 1;
  }
  else {             // n is negative
    l = -n;
    sign = -1;
  }
  lp = &l;
  ic = (unsigned char*)lp;
  while(ic[i] == 0 && --i >= 0);
  start = new unsigned char[size = i + 1];
  for(unsigned long j = 0; j < size; j++)
    start[j] = ic[j];
};

Z::Z(const unsigned long &n):
  sign(1) {
  short i = sizeof(long) - 1;
  const unsigned long *ip = &n;
  unsigned char *ic;
   ic = (unsigned char*)ip;
  while(ic[i] == 0 && --i >= 0);
  start = new unsigned char[size = i + 1];
  for(unsigned long j = 0; j < size; j++)
    start[j] = ic[j];
};

Z::Z(const string &s):
  size(0),
  start(new unsigned char[0]) {
  unsigned long begin = 0;
  Z factor = 1;
  char c;

  if (s[0] == '-')
    begin = 1;
  for (unsigned long i = s.length() - 1; i != begin - 1; i--, factor *= 10) {
    c = s[i];
    if (c >= '0' && c <= '9')
      *this += factor * (c - '0');
    else
      break;
  }
  sign = begin ? -1 : 1;
}

Z::Z(const Z &a):
  size(a.size),
  sign(a.sign) {
  start = new unsigned char[size];
  for (unsigned long i = 0; i < size; i++)
    start[i] = a.start[i];
}

// operators:

// addition and substraction operators:

Z operator+(const Z &a) {
  return a;
}

Z operator-(const Z &a) {
  Z temp = a;
  if (temp != 0)
    temp.sign = -temp.sign;
  return temp;
}

Z operator+(const Z &a, const Z &b) {
  Z temp = a;
  return temp += b;
}

Z operator-(const Z &a, const Z &b) {
  Z temp = a;
  return temp -= b;
}

Z& Z::operator++() {
  unsigned long i = 0;

  if (*this == -1) {
    start[0] = 0;
    sign     = 1;
  }
  else {
    if (*this >= 0) {
      while (start[i] == 255 && i < size)
        i++;
      if (i == size) {
        for (i = 0; i < size; i++)
          start[i] = 0;
        extend(1);
        start[i] = 1;
      }
      else {
        start[i]++;
        for(unsigned long j = 0; j < i; j++)
          start[j] = 0;
      }
    }
    else {
      while (start[i] == 0 && i < size)
        i++;
      start[i]--;
      if (i == size - 1 && start[i] == 0)
        shorten(1);
      for(unsigned long j = 0; j < i; j++)
        start[j] = 255;
    }
  }
  return *this;
}

Z Z::operator++(int) {
  Z temp = *this;
  ++*this;
  return temp;
}

Z& Z::operator--() {
  unsigned long i = 0;

  if (*this == 0) {
    start[0] =  1;
    sign     = -1;
  }
  else {
    if (*this > 0) {
      while (start[i] == 0 && i < size)
        i++;
      start[i]--;
      if (i == size - 1 && start[i] == 0)
        shorten(1);
      for(unsigned long j = 0; j < i; j++)
        start[j] = 255;
    }
    else {
      while (start[i] == 255 && i < size)
        i++;
      if (i == size) {
        for (i = 0; i < size; i++)
          start[i] = 0;
        extend(1);
        start[i] = 1;
      }
      else {
        start[i]++;
        for(unsigned long j = 0; j < i; j++)
          start[j] = 0;
      }
    }
  }
  return *this;
}

Z Z::operator--(int) {
  Z temp = *this;
  --*this;
  return temp;
}

Z Z::operator+=(const Z &a) {
  unsigned long mn = min(size, a.size),
                mx = max(size, a.size),
                shorter = a.size - size,
                i= 0;
  int sum, rest = 0;
  bool smaller = size < a.size;

  if (sign == a.sign) {
    for (; i < mn; i++) {
      sum = int(start[i]) + int(a.start[i]) + rest;
      start[i] = sum;
      rest = sum == start[i] ? 0 : 1;
    }
    if (smaller)
      extend(shorter);
    unsigned char *start2 = smaller ? a.start : start;
    for(; i < mx; i++) {
      sum = int(start2[i]) + rest;
      start[i] = sum;
      rest = sum == start[i] ? 0 : 1;
    }
    if (rest) {
      extend(1);
      start[size - 1] = rest;
    }
  }
  else {
    if (*this == -a)
      shorten(size);
    else {
      Z large, small;
      if (abs(*this) < abs(a)) {
        large = a     ;
        small = *this ;
        size  = a.size;
      }
      else {
        large = *this;
        small = a    ;
      }
      sign = large.sign;
      for (; i < mn; i++) {
        sum = large.start[i] - small.start[i] - rest;
        start[i] = sum;
        rest = sum < 0 ? 1 : 0;
      }
      for (; i < mx; i++) {
        sum = large.start[i] - rest;
        start[i] = sum;
        rest = sum < 0 ? 1 : 0;
      }
      shorten();
    }
  }
  return *this;
}

Z Z::operator-=(const Z &a) {
  return *this += -a;
}

// multiplication and division operators:

Z operator*(const Z &aa, const Z &bb) {
  if (aa.size == 0)
    return aa;
  if (bb.size == 0)
    return bb;
  Z a = aa, b = bb;
  short tempsign = a.sign * b.sign;
  unsigned long i = 0, j = 0, k = 0, d;
  unsigned char* temp;
  Z add, result;

  a.sign = 1;
  b.sign = 1;
  for (; i < a.size; i++)
    for (j = 0; j < b.size; j++) {
      d = i + j;
      add = (unsigned int) a.start[i] * (unsigned int) b.start[j];
      if (d > 0 && add != 0) {
        temp = add.start;
        add.start = new unsigned char[add.size += d];
        for (k = 0; k < d; k++)
          add.start[k] = 0;
        for (; k < add.size; k++)
          add.start[k] = temp[k - d];
        delete[] temp;
      }
      result += add;
    }
  result.sign = tempsign;
  return result;
}

Z operator/(const Z &aa, const Z &bb) {
  Z a = aa, b = bb;

  if (b == 0)
    Error(DivByZero, "Z/Z");
  if (abs(a) < abs(b))
    return Z(0);

  unsigned long len = a.size - b.size;
  short tempsign = a.sign * b.sign;
  unsigned int guess;
  Z aorg, c, result;

  delete[] result.start;
  result.start = new unsigned char[result.size = len + 1];
  for(unsigned long i = 0; i < result.size; i++)
    result.start[i] = 0;
  a.sign = 1;
  b.sign = 1;
  aorg = a;
  for (unsigned long i = len; i != -1lu; i--) {
    if (c.size != a.size - i) {
      delete[] c.start;
      c.start = new unsigned char[c.size = a.size - i];
    }
    for (unsigned long j = 0; j < c.size; j++)
      c.start[j] = a.start[i + j];
    if (b.size == c.size)
      guess = c.start[c.size - 1] / b.start[b.size - 1];
    else
      guess = (256 * c.start[c.size - 1] + c.start[c.size - 2]) / b.start[b.size - 1];
    while (b * guess > c)
      guess--;
    result.start[i] = guess;
    if (guess != 0)
      a = aorg - b * result;
    if (a.size == 0)
      break;
  }
  result.shorten();
  result.sign = tempsign;
  return result;
}

Z operator%(const Z &a, const Z &b) {
  Z c = a - a / b * b;
  return c < 0 ? c + b : c;
}

Z operator^(const Z &a, const Z &b) {
  Z temp = a;
  return temp ^= b;
}

Z Z::operator*=(const Z &a) {
  *this = *this * a;
  return *this;
}

Z Z::operator/=(const Z &a) {
  *this = *this / a;
  return *this;
}

Z Z::operator%=(const Z &a) {
  *this = *this % a;
  return *this;
}

Z Z::operator^=(Z a) {
  if (a < 0)
    Error(ExpLessZero, "Z^=Z");
  Z mult = *this, prod(1);
  for (;a != 0; a /= 2, mult *= mult)
    if (a % 2 == 1)
      prod *= mult;
  *this = prod;
  return prod;
}

// comparison operators:

bool operator==(const Z &a, const Z &b) {
  if (a.sign != b.sign || a.size != b.size)
    return false;
  for (unsigned long i = 0; i < a.size; i++)
    if (a.start[i] != b.start[i])
      return false;
  return true;
}

bool operator!=(const Z &a, const Z &b) {
  return !(a == b);
}

bool operator<(const Z &a, const Z &b) {
  if (a.sign == 1 && b.sign == -1)
    return false;
  if (a.sign == -1 && b.sign == 1)
    return true;
  if (a.size < b.size)
    return a.sign + 1;
  if (a.size > b.size)
    return a.sign - 1;
  for (unsigned long i = a.size - 1; i != -1ul; i--) {
    if (a.start[i] < b.start[i])
      return a.sign + 1;
    if (a.start[i] > b.start[i])
      return a.sign - 1;
  }
  return false;
}

bool operator>(const Z &a, const Z &b) {
  return b < a;
}

bool operator<=(const Z &a, const Z &b) {
  return a < b || a == b;
}

bool operator>=(const Z &a, const Z &b) {
  return b < a || a == b;
}

// assignement operator:

Z Z::operator=(const Z &a) {
  if (this != &a) {
    sign = a.sign;
    delete[] start;
    start = new unsigned char[size = a.size];
    for (unsigned long i = 0; i < size; i++)
      start[i] = a.start[i];
  }
  return *this;
}

// i/o operators:

ostream& operator<<(ostream &out, const Z &a) {
  out << string(a);
  return out;
}

istream& operator>>(istream &in, Z &a) {
  string s;
  in >> s;
  Z temp(s);
  a = temp;
  return in;
}

// type conversion:

Z::operator int() {
  unsigned short m = sizeof(int);
  int *result;
  unsigned long i = 0;
  char c[m];

  for(; i < m && i < size; i++)
    c[i] = start[i];
  for(; i < m; i++)
    c[i] = 0;
  result = (int*)c;
  return sign == 1 ? *result : -*result;
}

Z::operator int() const {
  unsigned short m = sizeof(int);
  int *result;
  unsigned long i = 0;
  char c[m];

  for(; i < m && i < size; i++)
    c[i] = start[i];
  for(; i < m; i++)
    c[i] = 0;
  result = (int*)c;
  return sign == 1 ? *result : -*result;
}

Z::operator unsigned int() {
  unsigned short m = sizeof(int);
  unsigned int *result;
  unsigned long i = 0;
  char c[m];

  for(; i < m && i < size; i++)
    c[i] = start[i];
  for(; i < m; i++)
    c[i] = 0;
  result = (unsigned int*)c;
  return sign == 1 ? *result : -*result;
}

Z::operator unsigned int() const {
  unsigned short m = sizeof(int);
  unsigned int *result;
  unsigned long i = 0;
  char c[m];

  for(; i < m && i < size; i++)
    c[i] = start[i];
  for(; i < m; i++)
    c[i] = 0;
  result = (unsigned int*)c;
  return sign == 1 ? *result : -*result;
}

Z::operator long() {
  unsigned short m = sizeof(long);
  unsigned long i = 0;
  long *result;
  char c[m];

  for(; i < m && i < size; i++)
    c[i] = start[i];
  for(; i < m; i++)
    c[i] = 0;
  result = (long*)c;
  return sign == 1 ? *result : -*result;
}

Z::operator long() const {
  unsigned short m = sizeof(long);
  unsigned long i = 0;
  long *result;
  char c[m];

  for(; i < m && i < size; i++)
    c[i] = start[i];
  for(; i < m; i++)
    c[i] = 0;
  result = (long*)c;
  return sign == 1 ? *result : -*result;
}

Z::operator unsigned long() {
  unsigned short m = sizeof(long);
  unsigned long i = 0, *result;
  char c[m];

  for(; i < m && i < size; i++)
    c[i] = start[i];
  for(; i < m; i++)
    c[i] = 0;
  result = (unsigned long*)c;
  return sign == 1 ? *result : -*result;
}

Z::operator unsigned long() const {
  unsigned short m = sizeof(long);
  unsigned long i = 0, *result;
  char c[m];

  for(; i < m && i < size; i++)
    c[i] = start[i];
  for(; i < m; i++)
    c[i] = 0;
  result = (unsigned long*)c;
  return sign == 1 ? *result : -*result;
}

Z::operator short() {
  unsigned short m = sizeof(short);
  short *result;
  unsigned long i = 0;
  char c[m];

  for(; i < m && i < size; i++)
    c[i] = start[i];
  for(; i < m; i++)
    c[i] = 0;
  result = (short*)c;
  return sign == 1 ? *result : -*result;
}

Z::operator short() const {
  unsigned short m = sizeof(short);
  short *result;
  unsigned long i = 0;
  char c[m];

  for(; i < m && i < size; i++)
    c[i] = start[i];
  for(; i < m; i++)
    c[i] = 0;
  result = (short*)c;
  return sign == 1 ? *result : -*result;
}

Z::operator unsigned short() {
  unsigned short m = sizeof(short), *result;
  unsigned long i = 0;
  char c[m];

  for(; i < m && i < size; i++)
    c[i] = start[i];
  for(; i < m; i++)
    c[i] = 0;
  result = (unsigned short*)c;
  return sign == 1 ? *result : -*result;
}

Z::operator unsigned short() const {
  unsigned short m = sizeof(short), *result;
  unsigned long i = 0;
  char c[m];

  for(; i < m && i < size; i++)
    c[i] = start[i];
  for(; i < m; i++)
    c[i] = 0;
  result = (unsigned short*)c;
  return sign == 1 ? *result : -*result;
}

Z::operator string() {
  string S;
  Z temp = *this;
  if (temp.size == 0) {
    S = '0';
    return S;
  }
  if (temp.sign == -1) {
    S = '-';
    temp.sign = 1;
  }
  short s;
  unsigned long m = 3 * temp.size, i = m - 1;
  char result[m + 1];
  result[m] = 0;
  for (; temp != 0; i--) {
    s = short(temp % 10);
    temp /= 10;
    result[i] = char(s) + '0';
  }
  S += result + i + 1;
  return S;
}

Z::operator string() const {
  Z temp = *this;
  return string(temp);
}

// methods:

string Z::classtype() {
  return "ring";
}

void Z::extend(const unsigned long &n) {
  unsigned long i = 0;
  unsigned char *temp = start;

  start = new unsigned char[size + n];
  for (; i < size; i++)
    start[i] = temp[i];
  size += n;
  for(; i < size; i++)
    start[i] = 0;
  delete[] temp;
}

void Z::shorten() {
  unsigned long n = size - 1;
  for (; n != -1lu; n--)
    if (start[n] != 0)
      break;
  shorten(size - n - 1);
}

void Z::shorten(const unsigned long n) {
  unsigned char *temp = start;
  unsigned long m = min(n, size);

  start = new unsigned char[size -= m];
  for(unsigned long i = 0; i < size; i++)
    start[i] = temp[i];
  delete[] temp;
  if (size == 0)
    sign = 1;
}

// functions:

Z abs(const Z &a) {
  Z temp = a;
  temp.sign = 1;
  return temp;
}

bool iseven(const Z &a) {
  return a % 2 == 0;
}

bool isodd (const Z &a) {
  return a % 2 == 1;
}

Z sqrt(const Z &a) {
  if (a < numeric_limits<unsigned long>::max())
    return (unsigned long)sqrt((unsigned long)a);
  else {

    // anyone who knows a better square root algorithm for positive numbers please tell me...

    Z result, temp = a;
    unsigned int first;
    unsigned long i = 0;

    if (a.size % 2 == 0) {
      result.size = a.size / 2;
      first = 256 * a.start[a.size - 1] + a.start[a.size - 2];
    }
    else {
      result.size = a.size / 2 + 1;
      first = a.start[a.size - 1];
    }
   for (; i < result.size - 1; i++)
      result.start[i] = 255;
    result.start[i] = int(sqrt(first));
    while (result * result > a)  //
      result--;                  // that is the bad part... =(
    return result;
  }
}

